/*
    Program: KEYTEST.PRG
    Author:  Greg Lief
    Copyright (c) 1991 Greg Lief
    Dialect: Clipper 5.01
    For Use by AQUARIUM readers (wild, wacky, wonderful)
    Compile with: clipper keytest /n
    Link with:    rtlink fi keytest /free

    If you want to use the @..GGET command in your own programs,
    copy the @..GGET directive to your own header file and then
    #include that header file at the top of your programs.
*/

// PROPER clause
#xcommand @ <row>, <col> GGET <var>                                     ;
                        [PICTURE <pic>]                                 ;
                        [VALID <valid>]                                 ;
                        [WHEN <when>]                                   ;
                        [PROPER]                                        ;
                        [MESSAGE <message>]                             ;
                        [COLOR <color>]                                 ;
                                                                        ;
      => SetPos( <row>, <col> )                                         ;
       ; AAdd(                                                          ;
               GetList,                                                 ;
               _GET_( <var>, <(var)>, <pic>, <{valid}>, <{when}>)       ;
             )                                                     ;    ;
       ATail(GetList):reader := { | g | MyReader(g,                     ;
       { | get, key | gkeyproper(get, key) } ) }                        ;
       [ ; Atail(GetList):colorDisp( <color> )]                         ;
       [ ; Atail(GetList):cargo := { <message> } ]

// STEP clause
#xcommand @ <row>, <col> GGET <var>                                     ;
                        [PICTURE <pic>]                                 ;
                        [VALID <valid>]                                 ;
                        [WHEN <when>]                                   ;
                        [STEP]                                          ;
                        [MESSAGE <message>]                             ;
                        [COLOR <color>]                                 ;
                                                                        ;
      => SetPos( <row>, <col> )                                         ;
       ; AAdd(                                                          ;
               GetList,                                                 ;
               _GET_( <var>, <(var)>, <pic>, <{valid}>, <{when}>)       ;
             )                                                     ;    ;
       ATail(GetList):reader := { | g | MyReader(g,                     ;
       { | get, key | gkeystep(get, key) } ) }                          ;
       [ ; Atail(GetList):colorDisp( <color> )]                         ;
       [ ; Atail(GetList):cargo := { <message> } ]


// PASSWORD clause, using default character ("*")
#xcommand @ <row>, <col> GGET <var>                                     ;
                        [PICTURE <pic>]                                 ;
                        [VALID <valid>]                                 ;
                        [WHEN <when>]                                   ;
                        [PASSWORD]                                      ;
                        [MESSAGE <message>]                             ;
                        [COLOR <color>]                                 ;
                                                                        ;
      => SetPos( <row>, <col> )                                         ;
       ; AAdd(                                                          ;
               GetList,                                                 ;
               _GET_( <var>, <(var)>, <pic>, <{valid}>, <{when}>)       ;
             )                                                     ;    ;
       ATail(GetList):reader := { | g | MyReader(g,                     ;
       { | get, key | gkeypass(get, key) } ) }                     ;    ;
       ATail(Getlist):picture := "@P"                                   ;
       [ ; Atail(GetList):colorDisp( <color> )]                         ;
       [ ; Atail(GetList):cargo := { <message> } ]


// MESSAGE clause only
#xcommand @ <row>, <col> GGET <var>                                  ;
                        [PICTURE <pic>]                              ;
                        [VALID <valid>]                              ;
                        [WHEN <when>]                                ;
                        [COLOR <color>]                              ;
                        [MESSAGE <message>]                          ;
                                                                     ;
      => SetPos( <row>, <col> )                                      ;
       ; AAdd(                                                       ;
               GetList,                                              ;
               _GET_( <var>, <(var)>, <pic>, <{valid}>, <{when}> )   ;
             )                                                     ; ;
       ATail(GetList):reader := { | g | MyReader(g) }                ;
       [ ; Atail(GetList):colorDisp( <color> )]                      ;
       [ ; Atail(GetList):cargo := <message> ]


#define TEST             // for compiling test code -- remove if desired


#ifdef TEST              // begin test code

function main
local w := space(20), x := space(20), y := 10, z := date()
memvar getlist
set(_SET_MESSAGE, 15)
setcolor('+gr/b')
cls
@ 10, 10 gget w password message "Please enter your password"
@ 11, 10 gget x proper   message "This will be proper-case"
@ 12, 10 gget y step     valid y < 30  ;
         message "Use the '+' and '-' keys, and keep it under 30"
@ 13, 10 gget z step     valid year(z) = 1991  ;
         message "Use the '+' and '-' keys, and keep it in 1991"
read
if ! empty(w)
   @ 16, 10 say "Your password was " + w
endif
return nil

#endif                   // end test code

/*------------------------------------------------------------------*/

#include "getexit.ch"    // necessary for the exitState constants
#include "inkey.ch"

#define  MESSAGE        1
#define  PASSWORD       cargo[2, 1]
#define  PASSWIDTH      cargo[2, 2]
#define  PASSCHAR       "*"

#xtranslate UsingPassword() => ;
            oGet:picture != NIL .and. substr(oGet:picture, 1, 2) == "@P"

/*
    Function: MyReader()
    Purpose:  Alternate to GetReader()
    Author:   Greg Lief
    Copyright (c) 1991 Greg Lief
    Dialect: Clipper 5.01
*/
function MyReader( oGet, bKeyproc )
local mess_row := set(_SET_MESSAGE)       // row for displaying messages

// read the GET if the WHEN condition is satisfied
if ( GetPreValidate(oGet) )

   // show message if there is one for this GET
   // note that current SET MESSAGE row will be used
   // if you prefer, you could default these to some other row
   if oGet:cargo != NIL .and. ! empty(oGet:cargo[MESSAGE])
      @ mess_row, 0 say padc( oGet:cargo[MESSAGE], maxcol() + 1)
   endif

   if UsingPassword()

      // if we have already been in this GET,
      // the cargo array will already contain 2 items
      if oGet:cargo != NIL .and. len(oGet:cargo) == 2
         oGet:varPut(padr(replicate(PASSCHAR, len(oGet:PASSWORD)), ;
                     oGet:PASSWIDTH))
      else
         // determine if cargo is an array or not
         if oGet:cargo == NIL
            oGet:cargo := { NIL }            // account for message slot
         endif

         // add subarray to cargo holding the following items:
         //    1) empty string to hold the contents of the password
         //    2) width of GET for padding later
         aadd(oGet:cargo, { '', len(oGet:varGet()) } )

      endif

   endif

   // activate the GET for reading
   oGet:SetFocus()

   // force cursor to end of the GET for password-style
   // if there is already data in this GET
   if UsingPassword() .and. ! empty(oGet:buffer)
      oGet:end()
   endif

   do while ( oGet:exitState == GE_NOEXIT )

      // check for initial typeout (no editable positions)
      if ( oGet:typeOut )
         oGet:exitState := GE_ENTER
      endif

      // apply keystrokes until exit
      do while ( oGet:exitState == GE_NOEXIT )
         eval(bKeyproc, oGet, inkey(0))
      enddo

      // disallow exit if the VALID condition is not satisfied
      if ( !GetPostValidate(oGet) )
         oGet:exitState := GE_NOEXIT
      endif

   enddo

   // remove message for this GET if there was one
   if oGet:cargo != NIL .and. ! empty(oGet:cargo[MESSAGE])
      scroll(mess_row, 0, mess_row, maxcol(), 0)
   endif

   // de-activate the GET
   oGet:KillFocus()

   // if password style entry was used, time to actually assign the
value
   if UsingPassword()
      oGet:varPut(padr(oGet:PASSWORD, oGet:PASSWIDTH))
   endif

endif

return nil

// end of function MyReader()


/*
    Function: GKeyBasic()
    Purpose:  test for hot keys and all basic keys used for exiting GETs
              This is a separate function so that the code need not be
              duplicated in each special keypress module.
    Author:   Greg Lief
    Copyright (c) 1991 Greg Lief
    Dialect:  Clipper 5.01
*/
function GKeyBasic( get, key )
local bKeyBlock := SetKey(key)
local ret_val := .t.
// check for SET KEY first
if bKeyBlock != NIL
   GetDoSetKey(bKeyBlock, get)
else
   do case

      case ( key == K_UP )
         get:exitState := GE_UP

      case ( key == K_SH_TAB )
         get:exitState := GE_UP

      case ( key == K_DOWN )
         get:exitState := GE_DOWN

      case ( key == K_TAB )
         get:exitState := GE_DOWN

      case ( key == K_ENTER )
         get:exitState := GE_ENTER

      case ( key == K_ESC )
         if ( Set(_SET_ESCAPE) )
            get:undo()
            get:exitState := GE_ESCAPE
         endif

      case ( key == K_PGUP )
         get:exitState := GE_WRITE

      case ( key == K_PGDN )
         get:exitState := GE_WRITE

      case ( key == K_CTRL_HOME )
         get:exitState := GE_TOP

      // both ^W and ^End terminate the READ (the default)
      case (key == K_CTRL_W)
         get:exitState := GE_WRITE

      otherwise
         ret_val := .f.
   endcase
endif
return (ret_val)

// end of function GKeyBasic()


/*
    Function: GKeyPass()
    Purpose:  Alternate for GetApplyKey() for password entry
    Author:   Greg Lief
    Copyright (c) 1991 Greg Lief
    Dialect:  Clipper 5.01
*/

// if you put this into a separate PRG, uncomment the next 5 lines
* #include "inkey.ch"
* #include "getexit.ch"
* #define  PASSWORD       cargo[2, 1]
* #define  PASSWIDTH      cargo[2, 2]
* #define  PASSCHAR       "*"


function gkeypass(oGet, key)
local cKey
local oldvalue

// check for SET KEY and basic movement keys
if ! GKeyBasic(oGet, key)

   do case

      case (key == K_INS)
         Set( _SET_INSERT, ! Set(_SET_INSERT) )
         setcursor( if(set(_SET_INSERT), 3, 1) )

      case (key == K_CTRL_U)
         oGet:Undo()

      case (key == K_BS)
         oGet:BackSpace()
         oGet:PASSWORD := substr(oGet:PASSWORD, 1, len(oGet:PASSWORD) - 1)

      otherwise

         if (key >= 32 .and. key <= 255)
            cKey := chr(key)

            // keep the real string in oGet:cargo and display
asterisks
            oGet:PASSWORD += ckey
            if ( Set(_SET_INSERT) )
               oGet:Insert(PASSCHAR)
            else
               oGet:Overstrike(PASSCHAR)
            endif
            if (oGet:typeOut .and. !Set(_SET_CONFIRM) )
               if ( Set(_SET_BELL) )
                  ?? Chr(7)
               endif
               oGet:exitState := GE_ENTER
            endif
         endif
   endcase
endif
return nil

// end of function GKeyPass()


/*
    Function: GKeyProper()
    Purpose:  Alternate for GetApplyKey() for proper-case data entry
    Author:   Greg Lief
    Copyright (c) 1991 Greg Lief
    Dialect:  Clipper 5.01
*/

// if you put this into a separate PRG, uncomment the next 2 lines
* #include "inkey.ch"
* #include "getexit.ch"

function gkeyproper(get, key)
static forceupper := .t.

local cKey
// check for SET KEY and basic movement keys
if ! GKeyBasic(get, key)
   do case
      case (key == K_INS)
         Set( _SET_INSERT, ! Set(_SET_INSERT) )
         setcursor( if(set(_SET_INSERT), 3, 1) )

      case (key == K_CTRL_U)
         get:Undo()
         forceupper := .t.        // reset flag for next time

      case (key == K_HOME)
         get:Home()
         forceupper := .t.

      case (key == K_END)
         get:End()
         forceupper := .f.

      case (key == K_RIGHT)
         get:Right()
         forceupper := (substr(get:buffer, get:pos - 1, 1) == " ")

      case (key == K_LEFT)
         get:Left()
         // if we just backspaced to the start of a word or the
string,
         // we must reset the flag to force next character upper-case
         forceupper := (get:pos==1 .or. substr(get:buffer, get:pos-1, 1) ==
" ")

      case (key == K_CTRL_RIGHT)
         get:WordRight()
         forceupper := .t.  // cause we are at the beginning of a word

      case (key == K_CTRL_LEFT)
         get:WordLeft()
         forceupper := .t.  // cause we are at the beginning of a word

      case (key == K_BS)
         get:BackSpace()
         /* if we just backspaced to the start of a word, or
            the start of the string, we must reset the flag to
            force the next character to be upper-case
         */
         forceupper := (get:pos==1 .or. substr(get:buffer, get:pos-1, 1) ==
" ")

      case (key == K_DEL)
         get:Delete()

      case (key == K_CTRL_T)
         get:DelWordRight()

      case (key == K_CTRL_Y)
         get:DelEnd()

      case (key == K_CTRL_BS)
         get:DelWordLeft()
         forceupper := .t.

      otherwise

         if (key >= 32 .and. key <= 255)
            cKey := Chr(key)
            if key == 32   /* space bar */
               forceupper := .t.  // next char will be uppercase
            elseif key > 64
               if forceupper
                  if key > 96 .and. key < 123
                     ckey := chr(key - 32)
                  endif
                  forceupper := .f.
               endif
            endif
            if ( Set(_SET_INSERT) )
               get:Insert(cKey)
            else
               get:Overstrike(cKey)
            endif
            if (get:typeOut .and. !Set(_SET_CONFIRM) )
               if ( Set(_SET_BELL) )
                  ?? Chr(7)
               endif
               get:exitState := GE_ENTER
            endif
         endif
   endcase
else
   forceupper := .t.
endif
return nil

// end of function GKeyProper()


/*
    Function: GKeyStep()
    Purpose:  Alternate for GetApplyKey() for step entry (dates and
numerics)
    Author:   Greg Lief
    Copyright (c) 1991 Greg Lief
    Dialect:  Clipper 5.01
*/

// if you put this into a separate PRG, uncomment the next 2 lines
* #include "inkey.ch"
* #include "getexit.ch"

function gkeystep(get, key)
local cKey
local oldvalue

// check for SET KEY and basic movement keys
if ! GKeyBasic(get, key)
   do case

      case (key == K_INS)
         Set( _SET_INSERT, ! Set(_SET_INSERT) )
         setcursor( if(set(_SET_INSERT), 3, 1) )

      case (key == K_CTRL_U)
         get:Undo()

      case (key == K_HOME)
         get:Home()

      case (key == K_END)
         get:End()

      case (key == K_RIGHT)
         get:Right()

      case (key == K_LEFT)
         get:Left()

      case (key == K_CTRL_RIGHT)
         get:WordRight()

      case (key == K_CTRL_LEFT)
         get:WordLeft()

      case (key == K_BS)
         get:BackSpace()

      case (key == K_DEL)
         get:Delete()

      case (key == K_CTRL_T)
         get:DelWordRight()

      case (key == K_CTRL_Y)
         get:DelEnd()

      case (key == K_CTRL_BS)
         get:DelWordLeft()

      otherwise

         if (key >= 32 .and. key <= 255)
            cKey := chr(key)
            // test for step entry on numerics and dates
            if cKey $ '-+' .and. get:type $ "ND"
               oldvalue := get:varGet()
               if cKey == "-"
                  get:varPut(get:varGet() - 1)
               else
                  get:varPut(get:varGet() + 1)
               endif
               if get:postBlock != NIL .and. ! eval(get:postBlock, get)
                  get:varPut(oldvalue)
               endif
               get:updateBuffer()
            else
               if ( Set(_SET_INSERT) )
                  get:Insert(cKey)
               else
                  get:Overstrike(cKey)
               endif
               if (get:typeOut .and. !Set(_SET_CONFIRM) )
                  if ( Set(_SET_BELL) )
                     ?? Chr(7)
                  endif
                  get:exitState := GE_ENTER
               endif
            endif
         endif
   endcase
endif
return nil

// end of function GKeyStep()

* eof GETS3.PRG
